import functools

from lhub.exceptions.validation import VersionMinimumNotMet


def minimum_version(min: float, feature_label: str):
    def min_version_decorator(func):
        functools.wraps(func)

        def minimum_version_wrapper(*args, **kwargs):
            version = float(args[0].version.lstrip("m"))
            if min > version:
                raise VersionMinimumNotMet(
                    min_version=f"m{min}",
                    feature_label=feature_label
                )
            return func(*args, **kwargs)

        return minimum_version_wrapper
    return min_version_decorator



import atexit
import json
import re
from collections import namedtuple

from requests import request, models
from requests.exceptions import HTTPError
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning
import sys
import time

from . import exceptions
from .log import prep_generic_logger
from .url import URLs
from .common import helpers
from logging import getLogger, RootLogger
from typing import List

log = getLogger(__name__)
cached_obj = namedtuple('CachedObject', ['time', 'value'])


class LogicHubAPI:
    DEFAULT_ALERT_RESULT_LIMIT = 100
    DEFAULT_CASE_RESULT_LIMIT = 25
    DEFAULT_CACHE_SECONDS = 300
    USER_AGENT_STRING = "LHUB CLI"
    http_timeout_default = 120
    verify_ssl = True
    last_response_status = None
    last_response_text = None

    __exit_set = False
    __version = None
    __version_info = None
    __case_prefix = None
    __fields = None
    __notebooks = None
    __hostname = None
    __user_id = None
    __user_role = None

    # Variables used only for API token auth
    __api_key = None

    # Variables used only for password auth
    __username = None
    __password = None
    _http_timeout_login = 20
    __credentials = {}
    __session_cookie = None

    def __init__(self, hostname, api_key=None, username=None, password=None,
                 verify_ssl=True, cache_seconds=None, default_timeout=None,
                 logger: RootLogger = None, log_level=None, **kwargs):
        global log
        if logger:
            log = logger
        else:
            prep_generic_logger(level=log_level)
        if not hostname:
            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="No hostname provided")
        if not api_key and not (username and password):
            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="Must provide either an API key or a username and password")
        if api_key and password:
            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="Using both an API token and a password is not supported")
        self.__hostname = hostname
        self.__api_key = api_key or None
        self.__username = username or None
        self.__password = password or None

        self.cache_seconds = int(cache_seconds or LogicHubAPI.DEFAULT_CACHE_SECONDS)
        if self.cache_seconds is None or self.cache_seconds < 0:
            self.cache_seconds = 0

        if default_timeout:
            self.http_timeout_default = int(default_timeout)
        self._http_timeout_login = min(self._http_timeout_login, self.http_timeout_default)
        self.kwargs = kwargs

        if isinstance(verify_ssl, bool):
            self.verify_ssl = verify_ssl
        if not self.verify_ssl:
            # Disable certificate warnings
            log.debug("Disabling SSL warnings...")
            disable_warnings(InsecureRequestWarning)

        if self.__api_key:
            self.auth_type = 'token'
        else:
            self.auth_type = 'password'
            self.__credentials = {"email": self.__username, "password": self.__password}

        __init_version = self.kwargs.pop('init_version', None)
        self.url = URLs(hostname, init_version=__init_version)
        log.debug(f"Starting session for server: {self.session_hostname}")
        self.__set_version(__init_version)
        self.formatted = FormattedObjects(api=self)

    def __enter__(self):
        return self

    def __get_cached_object(self, var: cached_obj):
        """
        Special method for managing cacheable property objects.
        To use this, the property must store values using the "cached_obj" namedtuple. (See "version_info" getter and setter as an example.)

        :param var:
        :return:
        """
        if var:
            if var.time >= int(time.time()) - self.cache_seconds:
                return var.value

    @property
    def default_http_headers(self):
        headers = {"User-Agent": self.USER_AGENT_STRING}
        if self.__api_key:
            headers["X-Auth-Token"] = self.__api_key
        if self.__session_cookie:
            headers['Cookie'] = f'PLAY_SESSION={self.__session_cookie}'
        return headers

    @property
    def version_info(self):
        if not self.__get_cached_object(self.__version_info):
            _ = self.get_version_info()
        return self.__version_info.value

    def __set_version_info(self, value):
        self.__version_info = cached_obj(int(time.time()), value)

    @property
    def fields(self):
        if not self.__get_cached_object(self.__fields):
            _ = self.list_fields()
        return self.__fields.value

    @fields.setter
    def fields(self, value):
        self.__fields = cached_obj(int(time.time()), value)

    @property
    def version(self):
        if not self.__get_cached_object(self.__version):
            _ = self.get_version_info()
        return self.__version.value

    @property
    def major_version(self):
        return int(float(self.version))

    @property
    def minor_version(self):
        return int(re.sub(r'^.*?(\d+)$', '$1', self.version))

    def __set_version(self, value):
        if value is None:
            _ = self.get_version_info()
            return
        self.__version = cached_obj(int(time.time()), helpers.format_version(value))
        log.debug(f"LogicHub version: {self.version}")
        # Update the version attribute in self.url for use in other calls since the URLs class does not make any such calls on its own
        self.url._current_version = self.version

    @property
    def case_prefix(self):
        if not self.__get_cached_object(self.__case_prefix):
            _ = self.cases_get_prefix()
        return self.__case_prefix.value

    @property
    def session_hostname(self):
        return self.__hostname

    @property
    def session_username(self):
        if not self.__username:
            _ = self.me()
        return self.__username

    @property
    def session_user_id(self):
        if not self.__user_id:
            _ = self.me()
        return self.__user_id

    @property
    def session_user_role(self):
        if not self.__user_role:
            _ = self.me()
        return self.__user_role

    @case_prefix.setter
    def case_prefix(self, value):
        self.__case_prefix = cached_obj(int(time.time()), value)

    @property
    def notebooks(self):
        if not self.__get_cached_object(self.__notebooks):
            result = self.list_notebooks()
            self.notebooks = result["result"]["data"]["data"]
        return self.__notebooks.value

    @notebooks.setter
    def notebooks(self, val):
        self.__notebooks = cached_obj(int(time.time()), val)

    @property
    def notebooks_name_map(self):
        return {notebook['name']: notebook['id']['id'] for notebook in self.notebooks}

    @property
    def session_cookie(self):
        # This should only ever come up w/ password auth, so if this has been invoked when there is an API key then throw an exception
        if self.__api_key:
            raise exceptions.LhBaseException('Session cookie should not be invoked')
        if not self.__session_cookie:
            self.login()
        return self.__session_cookie

    @session_cookie.setter
    def session_cookie(self, value):
        self.__session_cookie = value

    @property
    def cookies(self):
        if not self.__session_cookie:
            return {}
        return {'PLAY_SESSION': self.session_cookie}

    def __standard_http_response_tests(self, response_obj: models.Response, url: str, input_var=None):
        # If the response is okay, return successfully right away
        try:
            response_obj.raise_for_status()
            return
        except HTTPError as e:
            caught_exception = e

        try:
            __response_json = response_obj.json()
        except json.decoder.JSONDecodeError:
            __response_json = {}

        __status_code = response_obj.status_code
        __errors = __response_json.get("errors", [])
        __primary_error_type = __errors[0].get("errorType") if __errors else None
        exception_inputs = {
            "input_var": input_var,
            "url": url,
            "last_response_status": __status_code,
            "last_response_text": response_obj.text,
        }

        # Group all tests for status code 401
        if __status_code == 401:
            if self.auth_type == 'password':
                raise exceptions.auth.PasswordAuthFailure(**exception_inputs)
            else:
                if 'Unauthorized for URL' in response_obj.text or 'token is not allowed with this endpoint' in response_obj.text:
                    raise exceptions.auth.APIAuthNotAuthorized(url)
                else:
                    raise exceptions.auth.APIAuthFailure(url)

        # Group all tests for status code 400
        elif __status_code == 400:
            if self.auth_type == 'password':
                if url == self.url.login and 'Username/Password not valid' in self.last_response_text:
                    raise exceptions.auth.PasswordAuthFailure
            if 'batch' in url and re.search(r'controllers.BatchController.legacyPost.*?For input string', response_obj.text):
                _id = re.search(r'POST /demo/batch-(\d+)', response_obj.text)
                _id = _id.groups()[0] if _id else None
                raise exceptions.app.BatchNotFound(_id)
            if __primary_error_type == "AlreadyPresentException":
                if url == self.url.user_create:
                    raise exceptions.app.UserAlreadyExists(user=input_var)
                if url == self.url.user_group_create:
                    raise exceptions.app.UserGroupAlreadyExists(input_var=input_var)

        # Group all tests for status code 500
        elif __status_code == 500:
            if 'Unable to find batch with id' in response_obj.text:
                _id = re.search(r'Unable to find batch with id batch-(\d+)', response_obj.text)
                _id = _id.groups()[0] if _id else None
                raise exceptions.app.BatchNotFound(_id)

        raise caught_exception

    def _http_request(
            self, url, method="GET", params: dict = None, body=None,
            headers: dict = None, timeout=None, test_response: bool = True,
            reauth=True, input_var=None, **kwargs):
        method = method.upper() if method else "GET"
        # Only use reauth/automatic login if password auth is used
        if self.auth_type == 'password':
            # If called before successful login, run login process, but only if reauth was not disabled
            if reauth and not self.__session_cookie:
                self.login(force_new_session=True)
                # Disable re-authentication going forward
                reauth = False

        # Reset last response text
        self.last_response_text = self.last_response_status = None
        # Merge provided headers (if any) with default headers
        headers = {**self.default_http_headers, **(headers if headers else {})}
        timeout = timeout if timeout is not None else self.http_timeout_default
        json_body = None
        if isinstance(body, (dict, list)):
            json_body = body or {}
            body = None

        if method in ("POST", "PATCH") and not body and not json_body:
            body = "{}"
            json_body = {}

        kwargs = {k: v for k, v in {**{"params": params, "data": body, "json": json_body, "headers": headers}, **(kwargs or {})}.items() if v}
        response = request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)

        # Only use reauth/automatic login if password auth is used
        if self.auth_type == 'password':
            # if the response fails because the session has ended or the user has logged out, log back in and try once more
            if response.status_code == 401 and reauth and "not logged in" in response.text:
                log.debug(f"Session expired; logging in and retrying")
                self.login(force_new_session=True)
                response = request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)

        # Store last response before testing whether the call was successful
        self.last_url = url
        self.last_response_status = response.status_code
        self.last_response_text = response.text
        if test_response:
            self.__standard_http_response_tests(response, url, input_var=input_var)
        return response

    def login(self, force_new_session=False):
        if self.auth_type != 'password':
            raise exceptions.LhBaseException('Login process is only used for password auth')
        if force_new_session:
            self.__session_cookie = None

        if self.__session_cookie:
            log.debug("Already logged in; login skipped")
            return

        log.debug("Logging into LogicHub")
        login_response = self._http_request(url=self.url.login, method="POST", body=self.__credentials, timeout=self._http_timeout_login, reauth=False, input_var=self.__username)
        self.session_cookie = login_response.cookies.get("PLAY_SESSION")
        log.debug("Login successful")
        _ = atexit.register(self.close)

    def logout(self):
        log.debug(f"Logout requested [{self.session_hostname}]")
        if self.auth_type != 'password':
            raise exceptions.LhBaseException('Login process is only used for password auth')
        if not self.session_cookie:
            log.debug("Already logged out; logout skipped")
            return

        log.debug("Issuing logout command")
        _response = self._http_request(url=self.url.logout, method="POST", test_response=False, reauth=False, timeout=self._http_timeout_login, input_var=self.__username)
        try:
            _response.raise_for_status()
        except HTTPError as err:
            log.warning(f"Logout failed with HTTP error: {str(err)}")
        except Exception as err:
            log.warning(f"Logout failed with UNKNOWN error: {repr(err)}")
        else:
            log.debug(f"Logout successful")
            self.session_cookie = None

    def close(self):
        """Steps to clean up on exit"""
        if self.auth_type == 'password':
            log.debug("LogicHubAPI class exiting...")
            self.logout()

    # ToDo Revisit and finish this
    # def get_alerts(self, advanced_filter=None, limit=100):
    #     limit = int(limit) if limit is not None else 100_000
    #     body = {
    #         "pageNumber": 0,
    #         "pageSize": limit,
    #         "query": advanced_filter,
    #         "sortCol": "id",
    #         "sortOrder": "desc"
    #     }
    #     connection_list = []
    #     if not isinstance(connections, (list, tuple)):
    #         connections = [connections]
    #     for connection in connections:
    #         if isinstance(connection, str) and connection.startswith("connection-"):
    #             connection_list.append(connection)
    #         else:
    #             connection_list.append(f"connection-{connection}")
    #
    #     headers = {"Content-Type": "application/json"}
    #     # response = self._http_request(method="POST", url=self.url.connection_status, headers=headers, body=json.dumps(connection_list))
    #     response = self._http_request(method="POST", url=self.url.connection_status, headers=headers, body=connection_list)
    #     return response.json()

    def get_batch_results_by_id(self, batch_id: int, limit=1000, offset=0):
        """

        :param batch_id:
        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        limit = int(limit or 1000)
        offset = int(offset or 0)
        params = {"fields": "*", "pageSize": limit, "after": offset, "cachedOnly": "true"}
        log.debug("Fetching batch results")
        response = self._http_request(url=self.url.batch_results_by_id.format(int(batch_id)), params=params, input_var=batch_id)
        return response.json()

    def get_batches_by_stream_id(self, stream_id, limit=None, offset=0, statuses=None, exclude_empty_results=False):
        """

        :param stream_id:
        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :param statuses:
        :param exclude_empty_results:
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {
            "pageSize": limit,
            "after": int(offset or 0),
        }
        body = {"status": statuses or [], "excludeBatchesWithZeroEvents": exclude_empty_results}
        log.debug(f"Fetching batches for stream: {stream_id}")
        response = self._http_request(method="POST", url=self.url.stream_batches.format(stream_id), params=params, body=body, input_var=stream_id)
        return response.json()

    def get_connection_status(self, connections: list):
        """
        Get status of one or more connections
        :param connections: list of connection IDs, either as ints (like [6, 8]) or in full form (like ["connection-6", "connection-8"])
        :return:
        """
        connection_list = []
        if not isinstance(connections, (list, tuple)):
            connections = [connections]
        for connection in connections:
            if isinstance(connection, str) and connection.startswith("connection-"):
                connection_list.append(connection)
            else:
                connection_list.append(f"connection-{connection}")

        headers = {"Content-Type": "application/json"}
        log.debug("Fetching connection status")
        response = self._http_request(method="POST", url=self.url.connection_status, headers=headers, body=connection_list)
        return response.json()

    # ToDo in progress
    def get_custom_list_data(self, list_id, filter_sql: str = None, limit: int = None, offset: int = None):
        """

        :param list_id: numeric ID for requested custom list
        :param filter_sql: Optional: filter custom list data to return (spark SQL)
        :param limit: Optional: limit the number of results to return (default is 10,000)
        :param offset: Optional: if using pagination, provide the entry number to fetch (default is 0)
        :return:
        """
        assert isinstance(list_id, int) or int(list_id)
        list_id = int(list_id)
        if limit:
            assert isinstance(limit, int) or int(limit)
        if offset:
            assert isinstance(offset, int) or int(offset)
        limit = int(limit or 10_000)
        offset = int(offset or 0)
        if filter_sql:
            url = self.url.custom_list_data_with_filtering.format(list_id)
            method = "POST"
            payload = filter_sql.strip()
        else:
            url = self.url.custom_list_data
            method = "GET"
            payload = None
        log.debug("Fetching custom list data")
        response = self._http_request(
            method=method,
            url=url,
            params={"offset": offset, "limit": limit},
            body=payload,
            headers={'Content-Type': 'application/json'},
            input_var=list_id
        )
        results = response.json()
        try:
            results = results["result"]
        except Exception:
            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
        return results

    # ToDo Token auth not supported as of 2022-05-05 (m94)
    def get_dashboard(self, dashboard_id):
        log.debug("Fetching dashboard data")
        response = self._http_request(url=self.url.dashboard.format(dashboard_id), input_var=dashboard_id)
        return response.json()

    # ToDo Token auth not supported as of 2022-05-05 (m94)
    def get_dashboard_data(self, dashboard_id):
        log.debug("Fetching dashboard data")
        response = self._http_request(url=self.url.dashboard_data.format(dashboard_id), input_var=dashboard_id)
        return response.json()

    def get_playbook(self, resource_id):
        log.debug("Fetching playbook data")
        response = self._http_request(
            url=self.url.playbook.format(resource_id),
            body={"method": "getGraph", "parameters": {"filterByTag": None}},
            input_var=resource_id,
        )
        return response.json()

    # ToDo Token auth not supported as of 2022-05-05 (m94)
    def get_widget_data(self, dashboard_id, widget_id):
        log.debug("Fetching dashboard data")
        response = self._http_request(url=self.url.widget_data.format(dashboard_id=dashboard_id, widget_id=widget_id), input_var=widget_id)
        return response.json()

    # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
    def get_rule_set_by_name(self, name):
        rule_sets = self.list_rule_sets()
        rule_set = [x for x in rule_sets if x['name'] == name]
        if not rule_set:
            raise exceptions.app.RuleSetNotFound(input_var=name)
        rule_set = rule_set[0]
        rule_set['rules'] = self.get_rules_for_rule_set(rule_set['id'])
        return rule_set

    # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
    def get_rules_for_rule_set(self, rule_set):
        params = {
            "fields": "name,isPublic,rules[filter,values*1000,score]*1000",
        }
        log.debug("Fetching rules")
        response = self._http_request(url=self.url.rule_set.format(helpers.format_rule_set_id(rule_set)), params=params, input_var=rule_set)
        results = response.json()
        try:
            results = results["result"]["rules"]["data"]
        except Exception:
            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing rules of a rule set")
        return results

    def get_stream_by_id(self, stream_id: int):
        headers = {"Accept": "application/json"}
        log.debug("Fetching stream")
        response = self._http_request(url=self.url.stream_by_id.format(stream_id), headers=headers, test_response=False, input_var=stream_id)
        if response.status_code == 400 and 'Cannot find entity for id StreamId' in response.text:
            raise exceptions.app.StreamNotFound(stream_id)
        else:
            response.raise_for_status()
        return response.json()

    def get_version_info(self):
        log.debug("Fetching LogicHub version info")
        try:
            response = self._http_request(
                url=self.url.version,
                reauth=False,
                timeout=self.http_timeout_default)
            response_dict = response.json()
        except (KeyError, ValueError, TypeError, IndexError):
            raise exceptions.LhBaseException("LogicHub version could not be established")
        else:
            # Update version information any time this api call is run successfully
            self.__set_version_info(response_dict)
            self.__set_version(self.version_info)
        return response_dict

    def get_workflow_by_id(self, workflow_id: int):
        assert isinstance(workflow_id, int), "Workflow ID must be an integer"
        log.debug("Fetching workflow")
        response = self._http_request(method="GET", url=self.url.case_status_workflow_by_id.format(workflow_id), input_var=workflow_id)
        return response.json()

    # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
    # ToDo Create an action method for this, and update the beta integration action to use it
    def add_scoring_rule(self, rule_set, field_mappings, score: float or str):
        rule_set = helpers.format_rule_set_id(rule_set)
        field_mappings = helpers.sanitize_input_rule_field_mappings(field_mappings)
        score = helpers.format_rule_score(score)
        headers = {"Content-Type": "application/json"}
        body = {
            "method": "addRule",
            "parameters": {
                "values": [field_mappings],
                "score": score
            }
        }
        log.debug("Issuing new scoring rule")
        response = self._http_request(url=self.url.rule_set.format(rule_set), method="POST", body=body, headers=headers, input_var=rule_set)
        results = response.json()
        try:
            results = results["result"]
        except Exception:
            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for adding a scoring rule")
        return results

    def export_playbook(self, playbook_id, test_response=True):
        log.debug(f"Exporting playbook {playbook_id}")
        response = self._http_request(url=self.url.flow_export.format(playbook_id), test_response=test_response, input_var=playbook_id)
        return response.json()

    def execute_command(self, command_payload, limit=None):
        """

        :param command_payload:
        :param limit: None by default, although the LogicHub UI defaults to 25
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        log.debug("Executing LogicHub command")
        response = self._http_request(
            url=self.url.command_execute,
            method="POST",
            body=command_payload,
            test_response=False,
            params={"pageSize": limit}
        )
        try:
            result_dict = response.json()
        except json.decoder.JSONDecodeError:
            msg = f"Failed to load API response as JSON. Status code: {response.status_code} Response text: {response.text}"
            log.fatal(msg)
            raise exceptions.app.UnexpectedOutput(msg)

        errors = result_dict.pop("errors", [])
        if errors:
            log.warning(f"Request was successful, but command failed with errors")
            log.debug(f"Full error list: {json.dumps(errors)}")
            log.debug(f"Other error response data: {json.dumps(result_dict)}")
            try:
                log.warning(f"Direct link to command: {self.url.command.format(errors[0]['details']['flowId'])}")
            except KeyError:
                pass
            for error in errors:
                error_type = error.pop("errorType", None)
                error_msg = error.pop("message", None)
                error_details = json.dumps(error.pop("details", None))
                if error_details:
                    # Example error output:
                    # [DEBUG] [0x10f00cef0] Command error details: {"msg": "[Execution Error] Executing node final_output failed. [Error Executing Operator: select].
                    #   <<(lhub_file_id,None)>> does not exist in the table.", "flowId": "flow-1019", "nodeId": "ac4ce79f-d6d4-40fc-b855-77086fca92d5",
                    #   "causes": ["lql.LqlValidator$OperatorException: [Error Executing Operator: select]", "java.lang.IllegalArgumentException:
                    #   <<(lhub_file_id,None)>> does not exist in the table."]}
                    log.debug(f"Command error details: {error_details}")

                if error:
                    # [DEBUG] [0x10f00cef0] Additional error fields: {"cause": "lql.LqlValidator$OperatorException: [Error Executing Operator: select]"}
                    log.debug(f"Additional error fields: {json.dumps(error)}")

                # [FATAL] [0x11dc8b400] Command returned error (SparkRuntimeException): [Execution Error] Executing node final_output failed.
                # [Error Executing Operator: select]. <<(lhub_file_id,None)>> does not exist in the table.
                log.error(f"Command returned error ({error_type}): {error_msg}")
            sys.exit(1)

        try:
            response.raise_for_status()
        except HTTPError:
            message = f"Request failed with status code {response.status_code}\n\nText:\n\n"
            try:
                message += json.dumps(response.json(), indent=2)
            except (json.decoder.JSONDecodeError, ValueError, TypeError):
                message += response.text
            log.fatal(message)
        return response.json()

    def list_baselines(self, limit=None, offset=0, filters=None):
        """
        List all Baselines

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :param filters: Optional: Advanced search filters (list of dicts)
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {"libraryView": "all"}
        body = {"filters": filters or [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
        log.debug("Fetching baselines")
        response = self._http_request(method="POST", url=self.url.baselines, params=params, body=body)
        return response.json()

    def list_case_types(self, limit=None, after=0, exclude_deprecated: bool = True):
        """
        List all case types

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param after: Used for pagination if you want to pull in chunks, this sets the result number to start after
        :param exclude_deprecated: Exclude deprecated case types (default: True)
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {
            "pageSize": limit,
            "excludeDeprecated": str(exclude_deprecated or True).lower(),
            "after": after or 0,
        }
        log.debug("Fetching case types")
        response = self._http_request(url=self.url.case_types, params=params)
        return response.json()

    def search_entities(self, entity_type: str, query: str = None, limit: int = None, offset: int = 0):
        """
        Search content entities

        :param entity_type: type of entity to search (such as "eventTypes" or "connections")
        :param query: Search query (default: all entities of the given type)
        :param limit: Result limit (default: None)
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        body = {
            "query": query or "*",
            "pageSize": int(limit if limit and limit > 0 else 999999999),
            "page": offset
        }
        log.debug(f"Searching entities of type: {entity_type}")
        response = self._http_request(method="POST", url=self.url.entities_search.format(entity_type), body=body)
        return response.json()

    def list_commands(self, limit=None, offset=0, filters=None):
        """
        List all Commands

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :param filters: Optional: Advanced search filters (list of dicts)
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {"libraryView": "all"}
        body = {"filters": filters or [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
        log.debug("Fetching commands")
        response = self._http_request(method="POST", url=self.url.commands, params=params, body=body)
        return response.json()

    def list_connections(self, filters=None, limit=None, offset=0):
        """
        List all connections

        :param filters: Optional: Advanced search filters (list of dicts)
        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        limit = int(limit if limit and int(limit) and limit >= 1 else 99999)
        params = {"libraryView": "all"}
        headers = {"Content-Type": "application/json"}
        body = {"filters": filters or [], "offset": offset, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
        log.debug("Fetching connections")
        response = self._http_request(method="POST", url=self.url.connections, headers=headers, params=params, body=body)
        return response.json()

    def list_custom_lists(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = None, verify_results=True):
        """

        :param search_text: Optional: Part or all of the custom list name to filter by
        :param filters: Optional: Advanced search, e.g. [{"searchText":"<case name>"}]
        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :param verify_results: Optional: set to false to return response JSON without verifying that the format is as expected
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        offset = int(offset or 0)
        filters = filters or []
        if search_text:
            filters.append({"searchText": search_text})

        log.debug("Fetching custom lists")
        response = self._http_request(
            method="POST",
            url=self.url.custom_lists,
            params={"libraryView": "all"},
            body={
                "filters": filters,
                "offset": offset,
                "pageSize": limit,
                "sortColumn": "name",
                "sortOrder": "ASC"
            },
            headers={'Content-Type': 'application/json'},
        )
        results = response.json()
        if verify_results:
            try:
                _ = results["result"]["data"]
            except Exception:
                raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
        return results

    # ToDo Token auth not supported as of 2022-03-09 (m92)
    def list_dashboards(self):
        log.debug("Fetching dashboards")
        response = self._http_request(url=self.url.dashboards)
        return response.json()

    def list_dashboards_with_widgets(self):
        log.debug("Fetching dashboards with widgets")
        response = self._http_request(url=self.url.dashboards_and_widgets)
        return response.json()

    def __list_event_types_v1(self, limit=None):
        """

        :param limit: None by default, although the LogicHub UI defaults to 25
        :return:
        """
        # Not sure when this one stopped working, but it worked somewhere in at least m66 and didn't work any more in m70
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {"limit": limit}
        log.debug("Fetching event types")
        response = self._http_request(url=self.url.event_types, params=params)
        results = response.json()
        if results.get("result"):
            results = results["result"]
        if results.get("data"):
            results = results["data"]
        return results

    def __list_event_types_v2(self, limit=None, offset=0):
        """

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {"libraryView": "all"}
        # Confirm the right term for "name" as the sort column and change this sort order
        body = {"filters": [], "pageSize": limit, "offset": offset or 0, "sortColumn": "lastUpdated", "sortOrder": "DESC"}
        log.debug("Fetching event types")
        response = self._http_request(method="POST", url=self.url.event_types, params=params, body=body)
        results = response.json()
        if results.get("result"):
            results = results["result"]
        while isinstance(results, dict) and results.get('data') is not None:
            results = results["data"]
        return results

    def list_event_types(self, limit=None):
        """

        :param limit: None by default, although the LogicHub UI defaults to 25
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        # ToDo Remove this old version when it's safe to assume no one still uses m70
        if self.major_version < 70:
            return self.__list_event_types_v1(limit=limit)
        return self.__list_event_types_v2(limit=limit)

    def list_fields(self, params: dict = None, **kwargs):
        params = params or {"systemFields": "true", "pageSize": 9999, "after": 0}
        log.debug("Fetching fields")
        response = self._http_request(method="GET", url=self.url.fields, params=params, **kwargs)
        self.fields = response.json()
        return response.json()

    def list_integrations(self):
        log.debug("Fetching integrations")
        response = self._http_request(method="GET", url=self.url.integrations)
        return response.json()

    def list_ml_models(self):
        log.debug("Fetching ML models")
        response = self._http_request(
            method="GET",
            url=self.url.ml_models,
            headers={"Content-Type": "application/json"},
        )
        return response.json()

    def list_modules(self):
        log.debug("Fetching modules")
        response = self._http_request(
            method="GET",
            url=self.url.modules,
            headers={"Content-Type": "application/json"}
        )
        return response.json()

    def list_notebooks(self, limit=None, offset=0, search_string=None):
        """
        List all Notebooks

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :param search_string:
        :return:
        """
        limit = limit if limit and isinstance(limit, int) else 99999
        body = {
            "filters": [],
            "offset": offset or 0,
            "pageSize": limit,
            "sortColumn": "name",
            "sortOrder": "ASC",
        }
        if search_string:
            body['filters'] = [{"searchText": search_string}]
        log.debug("Fetching notebooks")
        response = self._http_request(
            url=self.url.notebooks_list,
            method="POST",
            headers={"Content-Type": "application/json"},
            params={"libraryView": "all"},
            body=body
        )
        response = response.json()
        # Sort the results by notebook ID
        self.notebooks = response["result"]["data"]["data"] = helpers.sort_notebook_objects_by_id(response["result"]["data"]["data"])
        return response

    def list_playbooks(self, limit=None, offset=0):
        """
        List all Playbooks

        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        body = {"filters": [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
        log.debug("Fetching playbooks")
        response = self._http_request(
            url=self.url.playbook_list,
            method="POST",
            params={"pageSize": limit},
            body=body
        )
        return response.json()

    def get_playbook_versions(self, playbook_id, limit=25, offset=0):
        limit = int(limit if limit is not None and limit > 0 else 25)
        offset = offset or 0
        log.debug(f"Fetching up to {limit} versions for playbook {playbook_id}")
        response = self._http_request(self.url.playbook_versions.format(playbook_id), params={"offset": offset, "limit": limit})
        return response.json()

    # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
    def list_rule_sets(self, limit=None):
        """

        :param limit: None by default, although the LogicHub UI defaults to 25
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {
            "fields": "name,isPublic",
            "pageSize": limit,
        }
        log.debug("Fetching rule sets")
        response = self._http_request(url=self.url.rule_sets, params=params)
        results = response.json()
        try:
            results = results["result"]["data"]
        except Exception:
            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing rule sets")
        return results

    def list_saml_configs(self):
        log.debug("Fetching SSO configurations")
        response = self._http_request(url=self.url.saml_configs)
        return response.json()

    def list_stream_states(self, stream_ids: List[int]):
        new_stream_id_list = []
        for s in stream_ids:
            new_stream_id_list.append(f'stream-{s}')
        body = {"streams": new_stream_id_list}
        log.debug("Fetching stream states")
        response = self._http_request(method="POST", url=self.url.stream_states, body=body, input_var=stream_ids)
        return response.json()

    def list_streams(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = 0):
        """
        List all streams

        :param search_text:
        :param filters:
        :param limit: None by default, although the LogicHub UI defaults to 25
        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
        :return:
        """
        limit = int(limit if limit and limit > 0 else 999999999)
        params = {"libraryView": "all"}
        headers = {"Content-Type": "application/json"}

        filters = filters or []
        if search_text:
            filters.append({"searchText": search_text})

        body_dict = {
            "filters": filters,
            "offset": offset or 0,
            "pageSize": limit,
            "sortColumn": "name",
            "sortOrder": "ASC"
        }

        log.debug("Fetching streams")
        response = self._http_request(method="POST", url=self.url.streams, body=body_dict, headers=headers, params=params)
        return response.json()

    def list_user_groups(self, limit=None, hide_inactive=True):
        limit = limit if limit and isinstance(limit, int) else 99999
        params = {"pageSize": limit, "after": 0}
        body = {"filters": []}
        if hide_inactive:
            body['filters'].append({"hideDeleted": True})
        log.debug("Fetching user groups")
        response = self._http_request(
            url=self.url.user_groups,
            method="POST",
            headers={"Content-Type": "application/json"},
            params=params,
            body=body
        )
        return response.json()

    def __list_users_v1_up_to_m95(self, limit=None, hide_inactive=True):
        limit = limit if limit and isinstance(limit, int) else 99999
        params = {"pageSize": limit, "after": 0}
        body = {"filters": []}
        if hide_inactive:
            body['filters'].append({"hideInactive": True})
        log.debug("Fetching users")
        response = self._http_request(
            url=self.url.users,
            method="POST",
            headers={"Content-Type": "application/json"},
            params=params,
            body=body
        )
        return response.json()

    def __list_users_v2_from_m96(self, limit=None, **kwargs):
        limit = limit if limit and isinstance(limit, int) else 99999
        params = {"pageSize": limit, "after": 0}
        body = {"filters": []}
        if kwargs:
            body["filters"].append(kwargs)
        log.debug("Fetching users")
        response = self._http_request(
            url=self.url.users,
            method="POST",
            headers={"Content-Type": "application/json"},
            params=params,
            body=body
        )
        return response.json()

    def list_users(self, limit=None, **kwargs):
        if self.major_version > 95 and "hide_inactive" in kwargs:
            raise DeprecationWarning("The \"hideInactive\" filter is no longer valid since LogicHub version m96")

        if self.major_version <= 95:
            func = self.__list_users_v1_up_to_m95
        else:
            func = self.__list_users_v2_from_m96
        return func(limit=limit, **kwargs)

    def list_workflows(self):
        log.debug("Fetching workflows")
        response = self._http_request(method="GET", url=self.url.case_status_list_workflows)
        return response.json()

    def reprocess_batch(self, batch_id):
        """
        Reprocess an errored, canceled, or stale batch using its batch ID

        :param batch_id:
        :return:
        """
        if str(batch_id).startswith("batch"):
            batch_id = int(re.sub(r'\D+', '', str(batch_id)))
        headers = {"Content-Type": "application/json"}
        body = {"method": "reschedule", "parameters": {}}
        log.debug("Issuing command to reprocess batch")
        response = self._http_request(method="POST", url=self.url.batch_reprocess.format(batch_id), headers=headers, body=body, input_var=batch_id)
        return response.json()

    # introduced somewhere between m70 and m80
    def me(self):
        log.debug("Fetching current user info")
        response = self._http_request(method="GET", url=self.url.me, timeout=30)
        result = response.json()
        if not all((
            result.get('result'),
            result['result']['id'],
            result['result']['username'],
            result['result']['role'],
            result['result']['preferences']
        )):
            raise exceptions.validation.ResponseValidationError(input_var=response, message="API response does not match the expected schema for user profile and preferences")

        self.__user_role = result['result']['role']
        if result['result']:
            new_id = result['result']['id']
            if isinstance(new_id, str):
                new_id = int(re.sub(r'\D+', '', new_id))
            if new_id != self.__user_id:
                self.__user_id = new_id
            # Username will only exist already if password auth is used. In case of API token auth, capture and record the username.
            if self.__username != result['result']['username']:
                self.__username = result['result']['username']
                log.debug(f"Current user updated to {self.__username} [ID: {self.__user_id}, role: {self.__user_role}]")
        return result

    def update_user(self, user_id, change_note=None, **user_kwargs):
        if not user_kwargs:
            raise exceptions.validation.InputValidationError("User update requested, but no changes provided")
        log.debug("Patching user" + (f" ({change_note})" if change_note else ""))
        response = self._http_request(method="PATCH", url=self.url.user.format(user_id), body=user_kwargs, input_var=user_id)
        return response.json()

    def update_current_user_preferences(self, preferences):
        return self.update_user(user_id=self.session_user_id, change_note="updating preferences", preferences=preferences)

    def case_prefix_refresh(self):
        _ = self.cases_get_prefix()
        return self.case_prefix

    # ToDo Revisit eventually. Aborted on 2022-02-22 because it turns out this
    #  can be set w/ the case management integration as a custom field, but
    #  there is still value in finishing this.
    # def case_link_alerts(self, case_id, alert_ids):
    #     if not isinstance(alert_ids, list):
    #         alert_ids = [alert_ids]
    #     updated_alerts = []
    #     for i in alert_ids:
    #         i = helpers.format_alert_id(i)
    #         if i:
    #             updated_alerts.append(i)
    #     alert_ids = updated_alerts
    #
    #     case_id = helpers.format_case_id_with_prefix(case_id, self.case_prefix)
    #
    #     field_id = self.formatted.system_field_lh_linked_alerts.get('id')
    #     body = {"fields": [{"id": field_id, "value": alert_ids}]}
    #     log.debug("Linking alert to case")
    #     response = self._http_request(
    #         method="PATCH",
    #         url=self.url.case_update_linked_alerts.format(case_id=case_id),
    #         headers={"Content-Type": "application/json"},
    #         input_var=case_id,
    #     )
    #     output = response.json()

    def case_update(self, case_id: str, **kwargs):
        if not kwargs:
            raise exceptions.validation.InputValidationError("No changes provided for updating case", input_var=case_id, action_description="case update")
        log.debug("Updating case")
        response = self._http_request(
            method="PATCH",
            url=self.url.case.format(case_id=case_id),
            headers={"Content-Type": "application/json"},
            body=kwargs
        )
        return response.json()

    def cases_get_prefix(self):
        log.debug("Fetching case prefix")
        response = self._http_request(
            url=self.url.cases_get_prefix,
            headers={"Content-Type": "application/json"}
        )
        output = response.json()
        self.case_prefix = output['result']
        return output

    def case_list_attached_notebooks(self, case_id, results_only=False):
        body = {
            "id": helpers.format_case_id_with_prefix(case_id, self.case_prefix),
            "key": "case",
        }
        log.debug("Fetching attached notebooks")
        response = self._http_request(
            url=self.url.notebooks_attached,
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body,
            input_var=case_id
        )
        response = response.json()
        # Sort the results by notebook ID
        results = response['result'] = helpers.sort_notebook_objects_by_id(response['result'])
        if results_only:
            for n in range(len(results)):
                results[n]['id']['id'] = int(results[n]['id']['id'])
            return results
        return response

    def case_overwrite_attached_notebooks(self, case_id, notebooks):
        notebooks = helpers.format_notebook_ids(notebooks)
        body = {
            "notebookAttachmentIds": notebooks,
            "notebookAttachedEntityId": {
                "key": "case",
                "id": helpers.format_case_id_with_prefix(case_id, self.case_prefix),
            }
        }
        log.debug("Updating attached notebooks")
        response = self._http_request(
            url=self.url.notebooks_attach,
            method="PATCH",
            headers={"Content-Type": "application/json"},
            params={"libraryView": "all"},
            body=body,
            input_var=case_id
        )
        return response.json()

    # ToDo Update alert search and validation methods to match the updates I made here
    def cases_search_validate(self, query: str):
        if query:
            # First confirm that the query is valid
            body = {"queryType": "advanced", "query": query}
            log.debug("Validating search syntax")
            response = self._http_request(
                url=self.url.cases_search_validate,
                method="POST",
                headers={"Content-Type": "application/json"},
                body=body
            )
            response = response.json()
            is_invalid = response['result'].get('error')
            if is_invalid is None:
                raise exceptions.app.UnexpectedOutput('Unexpected response while validating query string.')
            elif is_invalid:
                raise exceptions.app.CaseQueryValidationError(response['result'].get('message'))

    # ToDo Update alert search and validation methods to match the updates I made here
    def case_search_advanced(self, query: str = None, limit: int = None, **kwargs):
        # Validate inputs
        if limit is None:
            limit = self.DEFAULT_CASE_RESULT_LIMIT
        elif limit == -1:
            limit = 999999
        elif not isinstance(limit, int):
            raise exceptions.validation.InputValidationError("Invalid input type for search limit", input_var=limit)

        body = {
            "pageNumber": 0,
            "pageSize": limit,
            "query": query or "",
            "includeTaskEntityType": True,
            "includeWorkflow": True
        }
        if kwargs:
            body.update(kwargs)
        log.debug("Executing case search")
        response = self._http_request(
            url=self.url.cases_search_advanced,
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body
        )
        return response.json()

    def alert_fetch(self, alert_id, return_raw=False, **kwargs):
        alert_id = helpers.format_alert_id(alert_id)
        url = self.url.alert_fetch.format(alert_id)
        # Send request, but disable immediate testing in order to do specialized testing first
        log.debug("Fetching alert")
        response = self._http_request(method="GET", url=url, test_response=False, **kwargs, input_var=alert_id)
        if return_raw:
            return response
        if response.status_code == 400 and re.search(r'Alert with id alert-\d+ doesn\'t exist', response.text):
            raise ValueError(f'No alert with ID {alert_id}')
        self.__standard_http_response_tests(response, url, input_var=alert_id)
        return response.json()

    def alerts_search_validate(self, query: str):
        if query:
            # First confirm that the query is valid
            body = {"queryType": "advanced", "query": query}
            log.debug("Validating search syntax")
            response = self._http_request(
                url=self.url.alerts_search_validate,
                method="POST",
                headers={"Content-Type": "application/json"},
                body=body
            )
            response = response.json()
            is_invalid = response['result'].get('error')
            if is_invalid is None:
                raise exceptions.app.UnexpectedOutput('Unexpected response while validating query string.')
            elif is_invalid:
                raise exceptions.app.AlertQueryValidationError(response['result'].get('message'))

    def alerts_search_advanced(self, query: str = None, limit: int = None, **kwargs):
        # Sanitize inputs
        limit = int(limit) if limit and int(limit) > 0 else self.DEFAULT_ALERT_RESULT_LIMIT
        query = query.strip() if query and query.strip() else ""
        self.alerts_search_validate(query)

        # Now prep the actual query
        body = {
            "pageNumber": 0,
            "pageSize": limit,
            "sortCol": "id",
            "sortOrder": "desc",
            "query": query
        }
        if kwargs:
            body.update(kwargs)
        log.debug("Searching alerts")
        response = self._http_request(
            url=self.url.alerts_search_advanced,
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body
        )
        return response.json()

    def stream_pause(self, stream_id, **kwargs):
        # Sanitize stream ID
        body = [f"stream-{helpers.format_stream_id(stream_id)}"]
        log.debug("Pausing stream")
        response = self._http_request(method="POST", url=self.url.stream_pause, body=body, **kwargs, input_var=stream_id)
        return response.json()

    def stream_resume(self, stream_id, **kwargs):
        # Sanitize stream ID
        body = [f"stream-{helpers.format_stream_id(stream_id)}"]
        log.debug("Resuming stream")
        response = self._http_request(method="POST", url=self.url.stream_resume, body=body, **kwargs, input_var=stream_id)
        return response.json()

    def create_user(self, username, email, authentication_type, group_ids: list):
        body = {
            "username": username,
            "email": email,
            "userGroupIds": group_ids,
            "authenticationType": authentication_type or "password"
        }
        log.debug(f"Creating user: {username}")
        response = self._http_request(method="POST", url=self.url.user_create, body=body, input_var=username)
        return response.json()

    def delete_user(self, user_ids: list):
        log.debug(f"Deleting users by id: {', '.join([str(i) for i in user_ids])}")
        response = self._http_request(method="POST", url=self.url.user_delete, body=user_ids, input_var=user_ids)
        return response.json()

    def create_user_group(self, name: str, users: list = None, **kwargs):
        body = {"name": name, "users": users or [], **kwargs}
        log.debug(f"Creating user: {name}")
        response = self._http_request(method="POST", url=self.url.user_group_create, body=body, input_var=name)
        return response.json()

    def update_user_group(self, group_id: int, name: str, users: list = None, **kwargs):
        body = {"name": name, "users": users or [], **kwargs}
        log.debug(f"Creating user: {name}")
        response = self._http_request(method="PATCH", url=self.url.user_group.format(group_id), body=body, input_var=name)
        return response.json()

    def get_password_settings(self):
        log.debug("Fetching password settings")
        response = self._http_request(url=self.url.get_password_settings)
        return response.json()

    def update_password_settings(self, settings_dict):
        log.debug("Fetching password settings")
        response = self._http_request(method="PATCH", url=self.url.update_password_settings, body=settings_dict)
        return response.json()

    def reset_password(self, user_id):
        log.debug(f"Resetting password for user ID: {user_id}")
        response = self._http_request(method="POST", url=self.url.user_legacy.format(user_id), body={"method": "regeneratePassword", "parameters": {}})
        return response.json()

    def get_user_by_id(self, user_id):
        log.debug(f"Fetching user detail from ID: {user_id}")
        response = self._http_request(url=self.url.user_legacy.format(user_id))
        return response.json()


class FormattedObjects:
    def __init__(self, api: LogicHubAPI):
        self.__api = api

    @property
    def version(self):
        return super(self.__api).version

    @property
    def system_field_lh_linked_alerts(self) -> dict:
        field = None
        for f in self.__api.fields['result']['data']:
            if f.get('fieldName') == 'lh_linked_alerts':
                field = f
                break
        return field
        # raise exceptions.validation.VersionMinimumNotMet(min_version='m86', feature_label='linked alerts')

    @property
    def system_field_lh_linked_alerts_id(self):
        return int(self.system_field_lh_linked_alerts['id'].replace('field-', ''))

    @property
    def user_groups(self):
        groups = self.__api.list_user_groups(limit=None, hide_inactive=True)['result']['data']
        # convert the group ID to an int, rename field to 'id', and stick back at the top of the dict
        return [dict(**{'id': helpers.format_user_group_id(groups[n])}, **groups[n]) for n in range(len(groups))]

    @property
    def user_groups_by_id(self):
        return {g['usersGroupId']: g for g in self.user_groups}

    @property
    def user_groups_by_name(self):
        return {g['name']: g for g in self.user_groups}

    @property
    def user_groups_simple(self):
        return [{k: v for k, v in g.items() if k != 'entityTypePermissions'} for g in self.user_groups]

    @property
    def user_groups_simple_by_id(self):
        return {g['id']: g for g in self.user_groups_simple}

    @property
    def user_groups_simple_by_name(self):
        return {g['name']: g for g in self.user_groups_simple}

    @property
    def users(self):
        if self.__api.major_version < 96:
            kwargs = {"hide_inactive": True}
        else:
            kwargs = {"inactiveUsers": "onlyActive"}
        response = self.__api.list_users(limit=None, **kwargs)
        return response['result']['data']

    @property
    def users_by_id(self):
        users = self.users
        return {helpers.format_user_id(u): u for u in users}

    @property
    def users_by_name(self):
        users = self.users
        return {u['name']: u for u in users}

    def get_username_by_id(self, user_id):
        user = self.__api.get_user_by_id(user_id)
        return user["result"]["username"]



from .base import LhBaseException


class BaseFormatError(LhBaseException, ValueError):
    """Base exception for format errors for LogicHub resource objects"""
    message = "Invalid format for LH object"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class InvalidAlertIdFormat(BaseFormatError):
    """Invalid format for alert ID"""
    message = 'Invalid format for alert ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidBatchIdFormat(BaseFormatError):
    """Invalid format for batch ID"""
    message = 'Invalid format for batch ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidCaseIdFormat(BaseFormatError):
    """Invalid format for case ID"""
    message = 'Invalid format for case ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidConnectionIdFormat(BaseFormatError):
    """Invalid format for connection ID"""
    message = 'Invalid format for connection ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidNotebookIdFormat(BaseFormatError):
    """Invalid format for notebook IDs"""
    message = 'Invalid format for notebook ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidPlaybookIdFormat(BaseFormatError):
    """Invalid format for notebook IDs"""
    message = 'Invalid format for playbook ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidRuleScore(BaseFormatError):
    """Invalid score for scoring rule"""
    message = 'Invalid score for scoring rule'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


# ToDo Add the expected format to the default message.
class InvalidRuleFormat(BaseFormatError):
    """Invalid format for scoring rule"""
    message = 'Invalid format for scoring rule field mappings.'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidRuleSetIdFormat(BaseFormatError):
    """Invalid format for rule set ID"""
    message = 'Invalid format for rule set ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidStreamIdFormat(BaseFormatError):
    """Invalid format for Stream IDs"""
    message = 'Invalid format for stream ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidUserIdFormat(BaseFormatError):
    """Invalid format for user IDs"""
    message = 'Invalid format for user ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidVersionFormat(BaseFormatError):
    """Invalid Format for LogicHub Version"""
    message = 'Invalid format for product version'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)


class InvalidWorkflowIdFormat(BaseFormatError):
    """Invalid Workflow ID"""
    message = 'Invalid format for workflow ID'

    def __init__(self, input_var, *args, **kwargs):
        super().__init__(input_var=input_var, *args, **kwargs)



import json
import re
from copy import deepcopy

from .api import LogicHubAPI
from .common import helpers
from .common.time import epoch_time_to_str
from .exceptions.app import BaseAppError, UserGroupNotFound, UserNotFound, UnexpectedOutput
from .exceptions.validation import InputValidationError, ResponseValidationError
from .exceptions.formatting import InvalidWorkflowIdFormat
from .log import prep_generic_logger
from logging import getLogger, RootLogger
from typing import Union

log = getLogger(__name__)


class Actions:

    def __init__(self, api: LogicHubAPI):
        self.__api = api

    @staticmethod
    def __enrich_alert_data(alert: dict, included_standard_fields=None, included_additional_fields=None):
        # Verify and sanitize inputs
        if not alert:
            return {}
        alert = deepcopy(alert)
        if not isinstance(included_standard_fields, list):
            included_standard_fields = [included_standard_fields] if included_standard_fields else []
        included_standard_fields = [f.strip() for f in included_standard_fields if f and f.strip()]
        if not isinstance(included_additional_fields, list):
            included_additional_fields = [included_additional_fields] if included_additional_fields else []
        included_additional_fields = [f.strip() for f in included_additional_fields if f and f.strip()]

        if not included_standard_fields and not included_additional_fields:
            return alert

        additional_fields = alert.pop('additionalFields', {})
        if isinstance(additional_fields, list):
            additional_fields = [f for f in additional_fields if not included_additional_fields or f.get('displayName') in included_additional_fields]
        else:
            additional_fields = {k: v for k, v in additional_fields.items() if not included_additional_fields or k in included_additional_fields}

        mapped_fields = alert.pop('mappedAlertFieldValues', [])
        if isinstance(mapped_fields, list):
            mapped_fields = [f for f in mapped_fields if not included_standard_fields or f.get('displayName') in included_standard_fields]
        else:
            mapped_fields = {k: v for k, v in mapped_fields.items() if not included_standard_fields or k in included_standard_fields}

        new_alert = {'id': alert.pop('id')}
        new_alert.update({k: v for k, v in alert.items() if not included_standard_fields or k in included_standard_fields})
        new_alert.update({
            'additionalFields': additional_fields,
            'mappedAlertFieldValues': mapped_fields,
        })
        return new_alert

    @staticmethod
    def _result_dict_has_schema(result_dict, *fields, raise_errors=False, action_description=None, accept_null=False):
        _dict = result_dict
        for field in fields:
            if field not in _dict.keys() or _dict.get(field) is None and not accept_null:
                if raise_errors is True:
                    action_description = f" [{action_description}]" if action_description else ''
                    err_msg = f"Response{action_description} did not match the expected JSON schema. Missing expected key: {field}"
                    raise ResponseValidationError(input_var=err_msg, message=action_description)
                return False
            _dict = _dict[field]
        return True

    @property
    def playbook_ids(self):
        response = self.list_playbooks()
        return {p["id"]["id"]: p["name"] for p in response}

    @staticmethod
    def __reformat_cmd_results(response_dict, drop_hidden_columns=True):
        full_result = response_dict.copy()
        result_raw = full_result["result"]
        warnings = full_result["result"].get("warnings")
        result_with_schema = result_raw["rows"]["data"]

        if drop_hidden_columns:
            result_output = [{k: v for k, v in _result['fields'].items() if k not in ["lhub_id", "lhub_page_num"]} for _result in result_with_schema]
        else:
            result_output = [{k: v for k, v in _result['fields'].items()} for _result in result_with_schema]

        for _warning in warnings:
            log.warning(f"Warning returned: {_warning}")
        return result_output, warnings

    def execute_command(self, command_name, input_dict, reformat=True, result_limit=None):
        response = self.__api.execute_command({"command": command_name, "parameterValues": input_dict, "limit": result_limit})
        if not reformat:
            return response

        response, _ = self.__reformat_cmd_results(response)
        return response

    @staticmethod
    def __reformat_alert_simple(alert: dict):
        if not alert:
            return {}
        alert = deepcopy(alert)

        # From this: {"caseFieldId": "field-17", "displayName": "Alert Context", "fieldType": "Text", "value": "mdr_test"}
        # To this: {"Alert Context": "mdr_test"}
        for k in alert.get('mappedAlertFieldValues', {}):
            if k.get('fieldType') == 'JSON':
                try:
                    k['value'] = json.loads(k['value'])
                except (ValueError, TypeError, json.decoder.JSONDecodeError):
                    pass
        alert['mappedAlertFieldValues'] = {x['displayName']: x['value'] for x in alert.get('mappedAlertFieldValues', {}) if x.get('displayName')}
        return alert

    def get_alert_by_id(self, alert_id, simple_format=False, included_standard_fields=None, included_additional_fields=None):
        result = self.__api.alert_fetch(alert_id)
        _ = self._result_dict_has_schema(result, "result", action_description="fetch alert", raise_errors=True)

        # Reformat alert and enrich as needed
        output = self.__enrich_alert_data(
            result['result'],
            included_standard_fields=included_standard_fields,
            included_additional_fields=included_additional_fields
        )

        # Simplify mapped and additional field output
        if simple_format:
            output = self.__reformat_alert_simple(output)
        return output

    def get_batch_results_by_id(self, batch_id: int, limit=1000, keep_additional_info=False):
        result = self.__api.get_batch_results_by_id(batch_id=batch_id, limit=limit)
        _ = self._result_dict_has_schema(result, "result", "data", raise_errors=True, action_description="get batch results by batch ID")
        result = result["result"]["data"]
        if not keep_additional_info:
            result = [r["columns"] for r in result]
        return result

    def get_batches_by_stream_id(self, stream_id: int, limit=None, statuses=None, exclude_empty_results=False):
        """
        Get all batches for a given stream

        :param stream_id: ID of the stream
        :param limit: Optional: set a batch limit (default it unlimited, despite the fact that the API's actual default is 25)
        :param statuses: Optional: List of statuses to include in the results
        :param exclude_empty_results: Optional: exclude successful batches which did not output any data
        :return: Batch results in the form of a list of dicts
        """
        result = self.__api.get_batches_by_stream_id(stream_id, limit=limit or -1, statuses=statuses, exclude_empty_results=exclude_empty_results)
        _ = self._result_dict_has_schema(result, "result", "data", raise_errors=True, action_description="get batches by stream ID")
        return result["result"]["data"]

    def get_case_prefix(self):
        return self.__api.case_prefix

    def get_connection_status_by_id(self, connection_ids: list):
        connection_ids = list(set([helpers.format_connection_id(c) for c in connection_ids]))
        status_results = self.__api.get_connection_status(connection_ids)
        statuses = {helpers.format_connection_id(status["connectionEntityId"]): status["status"] for status in status_results['result']}
        return statuses

    # ToDo Token auth not supported as of 2022-03-09 (m92)
    def get_dashboard_data(self, dashboard_id):
        result = self.__api.get_dashboard_data(dashboard_id.strip())
        _ = self._result_dict_has_schema(result, "result", "widgets", raise_errors=True, action_description="get dashboard data")
        result = result['result']['widgets']
        return result

    def get_notebooks_attached_to_case(self, case_id):
        response = self.__api.case_list_attached_notebooks(case_id)
        _ = self._result_dict_has_schema(response, "result", action_description="get notebooks from case", raise_errors=True)
        return response.get('result', [])

    def get_stream_by_id(self, stream_id: int):
        result = self.__api.get_stream_by_id(stream_id)
        return result["result"]

    def get_version(self):
        return self.__api.version_info

    def get_workflow_by_id(self, workflow_id):
        if isinstance(workflow_id, str):
            workflow_id = re.findall(r"(\d+)", workflow_id)
            if not workflow_id:
                raise InvalidWorkflowIdFormat(input_var=workflow_id)
            workflow_id = workflow_id[0]
        workflow_id = int(workflow_id)
        result = self.__api.get_workflow_by_id(workflow_id=workflow_id)
        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="get workflow by ID")
        return result["result"]

    def list_stream_states(self, stream_ids, include_recent_batches=True, return_as_simple_dict=False):
        stream_ids = helpers.format_stream_id(stream_ids)
        result = self.__api.list_stream_states(stream_ids=stream_ids)
        _ = self._result_dict_has_schema(result, "result", "streams", raise_errors=True, action_description="get stream states")
        results = result['result']['streams']
        if return_as_simple_dict:
            return {r["streamId"]: r["status"] for r in results}
        if not include_recent_batches:
            for n in range(len(results)):
                _stream_id = results[n]['streamId']
                if 'data' in results[n]:
                    del results[n]['data']
                else:
                    log.warning(f"Expected recent batches in 'data' key of stream state, but no such key was found for stream {_stream_id}")
        return results

    def list_baselines(self, verify_stream_states=False):
        result = self.__api.list_baselines()
        _ = self._result_dict_has_schema(result, "result", "data", "data", raise_errors=True, action_description="list baselines")
        result = result["result"]["data"]
        baselines = result['data']
        stream_ids = [int(b['id']['id']) for b in baselines]
        stream_state_response = self.__api.list_stream_states(stream_ids=stream_ids)
        _ = self._result_dict_has_schema(stream_state_response, "result", "streams", raise_errors=True, action_description="list stream states")
        state_map = {_stream['streamId']: _stream['status'] for _stream in stream_state_response['result']['streams']}
        for n in range(len(baselines)):
            b = baselines[n]
            _id = int(b['id']['id'])
            baselines[n]['baseline_config_status'] = state_map.get(f"stream-{b['id']['id']}")

        if verify_stream_states is True:
            self.__update_stream_results_with_status(result["data"])
        return result

    @staticmethod
    def __reformat_command_simple(result):
        result = {
            "name": result["name"],
            "id": int(result["id"]["id"]),
            "flowId": result["flowId"],
            "owner": result["owner"],
            "last_updated": epoch_time_to_str(result["lastUpdated"] / 1000),
            "command_status": result["commandStatus"],
        }
        return result

    def list_commands(self, filters: list = None, limit: int = None, simple_format=False):
        result = self.__api.list_commands(filters=filters, limit=limit, offset=0)
        _ = self._result_dict_has_schema(result, "result", "data", "data", action_description="list commands", raise_errors=True)
        results = result["result"]["data"]
        log.debug(f"{len(results)} command{'s' if len(results) != 1 else ''} found")
        if simple_format:
            results = [self.__reformat_command_simple(r) for r in results['data']]
        return results

    def list_case_types(self, limit: int = None, after: int = None, exclude_deprecated: bool = None):
        result = self.__api.list_case_types(limit=limit, after=after, exclude_deprecated=exclude_deprecated)
        _ = self._result_dict_has_schema(result, "result", "data", action_description="list case types", raise_errors=True)
        result = result["result"]["data"]
        return result

    def list_connections(self, filters=None, add_status=False):
        result = self.__api.list_connections(filters=filters)
        _ = self._result_dict_has_schema(result, "result", "data", action_description="list connections", raise_errors=True)
        result = result["result"]["data"]
        if not add_status:
            return result

        connections = result["data"]
        connection_ids = [connection["id"]["id"] for connection in connections]
        statuses = self.get_connection_status_by_id(connection_ids)
        for n in range(len(connections)):
            connection = connections[n]
            connection_id = int(connection["id"]["id"])
            connection["status"] = statuses.get(connection_id)
            if not connection["status"]:
                raise BaseAppError(f"Connection ID {connection_id} missing from status list")
        return result

    def list_custom_lists(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = None):
        results = self.__api.list_custom_lists(search_text=search_text, filters=filters, limit=limit, offset=offset, verify_results=False)
        warnings = []
        try:
            results = results["result"]["data"]
        except KeyError:
            warnings.append("API response does not match the expected schema for listing custom lists")
        return results, warnings

    # ToDo Token auth not supported as of 2022-03-09 (m92)
    def list_dashboards(self):
        result = self.__api.list_dashboards()
        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="list dashboards")
        result = result['result']
        return result

    def list_dashboards_with_widgets(self, include_not_imported=True):
        result = self.__api.list_dashboards_with_widgets()
        _ = self._result_dict_has_schema(result, "result", "objects", raise_errors=True, action_description="list dashboards with widgets")
        result = result['result']['objects']
        if not include_not_imported:
            result = [r for r in result if r['metadata']["contentRepoStatus"] != "Global"]
        return result

    def list_event_types(self, limit=None, wait_for_connection_status=False):
        results = self.__api.list_event_types(limit=limit)
        if not wait_for_connection_status:
            return results
        connection_ids = []
        for r in results:
            try:
                connection_ids.append(f'connection-{helpers.format_connection_id(r["origin"]["id"]["id"])}')
            except KeyError:
                pass
        statuses = self.get_connection_status_by_id(list(set(connection_ids)))
        for r in results:
            try:
                if r["origin"]["id"]["id"]:
                    r["origin"]["connectionStatus"] = r["eventTypeConnectionStatus"] = statuses[int(r["origin"]["id"]["id"])]
            except KeyError:
                pass
        return results

    def list_fields(self, map_mode=None):
        assert not map_mode or map_mode in ['id', 'name'], f'Invalid output format: {map_mode}'
        response = self.__api.fields
        _ = self._result_dict_has_schema(response, "result", "data", raise_errors=True, action_description="list fields")
        output = response["result"]["data"]
        if map_mode == "id":
            output = {f['id']: {k: v for k, v in f.items() if k != 'id'} for f in output}
        elif map_mode == "name":
            output = {f['fieldName']: {k: v for k, v in f.items() if k != 'fieldName'} for f in output}
        return output

    def list_integrations(self, name_filter: str = None, filter_type="contains"):
        filter_type = filter_type.lower()
        assert filter_type in ["equals", "contains"], f"Invalid filter type \"{filter_type}\""

        response = self.__api.list_integrations()
        _ = self._result_dict_has_schema(response, "result", "objects")
        result = response.pop("result")
        categories = result["categories"]
        results = result["objects"]

        if name_filter:
            if filter_type == "contains":
                results = [result for result in results if name_filter.lower() in result["resource"]["name"].lower()]
            else:
                new_result = {}
                for result in results:
                    if result["resource"]["name"].lower() == name_filter.lower():
                        new_result = result
                        break
                results = new_result
        return results, categories

    def list_modules(self, local_only=False):
        result = self.__api.list_modules()
        _ = self._result_dict_has_schema(result, "result", "objects", action_description="list modules", raise_errors=True)
        other_details = result["result"]
        modules = other_details.pop('objects')

        # If local_only is set to True, only return the modules with a contentRepoStatus of "Local"
        if local_only is True:
            modules = [x for x in modules if x['metadata']['contentRepoStatus'].lower() == 'local']

        return modules, other_details

    def list_notebooks(self):
        result = self.__api.list_notebooks()
        _ = self._result_dict_has_schema(result, "result", "data", "data", action_description="list notebooks", raise_errors=True)
        return result["result"]["data"]

    def list_playbooks(self, limit=None, map_mode=None):
        assert not map_mode or map_mode in ['id', 'name'], f'Invalid output format: {map_mode}'
        response = self.__api.list_playbooks(limit=limit)
        _ = self._result_dict_has_schema(response, "result", "data", "data")
        output = response["result"]["data"]["data"]
        if map_mode == "id":
            output = {f['id']['id']: {k: v for k, v in f.items() if k != 'id'} for f in output}
        elif map_mode == "name":
            output = {f['name']: {k: v for k, v in f.items() if k != 'name'} for f in output}
        return output

    def get_playbook(self, resource_id):
        response = self.__api.get_playbook(resource_id=resource_id)
        _ = self._result_dict_has_schema(response, "result")
        return response["result"]

    def get_playbook_versions(self, playbook_id, **kwargs):
        response = self.__api.get_playbook_versions(playbook_id=playbook_id, **kwargs)
        _ = self._result_dict_has_schema(response, "result", "data")
        return response["result"]["data"]

    def list_saml_configs(self):
        result = self.__api.list_saml_configs()
        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="list SAML configs")
        return result["result"]

    def __update_stream_results_with_status(self, results):
        id_list = helpers.format_stream_id(results)
        states = self.list_stream_states(stream_ids=id_list, include_recent_batches=False, return_as_simple_dict=True)
        for r in results:
            if not r.get('id') or not r['id'].get('id'):
                raise UnexpectedOutput(f"stream dict did not match expected format: {r}")
            _id_str = f"stream-{r['id']['id']}"
            _id_str = f"stream-{r['id']['id']}"
            if not states.get(_id_str):
                raise UnexpectedOutput(f"{_id_str} was returned in the stream search but was not present in state search results")
            r['streamStatus'] = states[_id_str]

    def list_streams(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = 0, verify_stream_states=False):
        """
        List all streams (or streams matching a search filter)

        :param search_text: Partial or full name of the streams to return
        :param filters: Other search filters (see the streams API)
        :param limit: Limit the number of results to return
        :param offset: For pagination, provide the page number of results to return
        :param verify_stream_states: The streams API does not wait for stream status
        calculation to complete unless status is part of the search filter. Enabling
        this option will fetch the true state for every stream returned before returning results.
        :param return_as_simple_dict: Return a simple dict instead, with stream IDs as keys and their respective statuses as values.
        :return:
        """
        result = self.__api.list_streams(search_text=search_text, filters=filters, limit=limit, offset=offset)
        # ToDo get accurate states with actions.list_stream_states
        # ToDo file a bug for the streams api returning incorrect states
        _ = self._result_dict_has_schema(result, "result", "data", "data", raise_errors=True, action_description="list streams")
        results = result["result"]["data"]["data"]
        if verify_stream_states is True:
            self.__update_stream_results_with_status(results)

        return results

    def list_user_groups(self, hide_inactive=True):
        result = self.__api.list_user_groups(hide_inactive=hide_inactive)
        _ = self._result_dict_has_schema(result, "result", "data", action_description="list user groups", raise_errors=True)
        return result["result"]

    @staticmethod
    def __reformat_user_simple(user: dict):
        groups = [g['name'] for g in user['groups'] if not g.get("isDeleted", False)]
        group_ids = [g['id'] for g in user['groups'] if not g.get("isDeleted", False)]
        user_attributes = {
            "username": user.get("name"),
            "is_admin": user["role"]["value"] == "admin",
            "groups": ', '.join(groups),
            "group_ids": group_ids,
            "email": user.get("email"),
            "is_deleted": user.get("isDeleted"),
            "is_enabled": user.get("isEnabled"),
            "auth_type": user.get("authenticationType"),
            "id": user.get("userId"),
        }
        # would it be beneficial to make a class object for these instead of returning a dict?
        return user_attributes

    def list_users(self, hide_inactive=True, simple_format=False, **filters):
        if not filters:
            if self.__api.major_version < 96:
                filters = {"hide_inactive": hide_inactive}
            else:
                filters["inactiveUsers"] = "all"
                if hide_inactive is True:
                    filters["inactiveUsers"] = "onlyActive"

        result = self.__api.list_users(**filters)
        _ = self._result_dict_has_schema(result, "result", "data", action_description="list users", raise_errors=True)
        results = result["result"]
        result_count = len(results['data'])
        log.debug(f"{result_count} user{'s' if result_count != 0 else ''} found")
        if simple_format:
            results = [self.__reformat_user_simple(result) for result in results['data']]
        return results

    def reset_password(self, user_id):
        user_id = helpers.format_user_id(user_id)
        username = self.__api.formatted.get_username_by_id(user_id)
        log.debug(f"Resetting password for user: {username}")
        result = self.__api.reset_password(user_id=user_id)
        _ = self._result_dict_has_schema(result, "result", "password", action_description="reset password", raise_errors=True)
        return result["result"]["password"]

    def list_workflows(self):
        result = self.__api.list_workflows()
        _ = self._result_dict_has_schema(result, "result", "data", action_description="list workflows", raise_errors=True)
        return result["result"]["data"]

    def reprocess_batch(self, batch_id):
        return self.__api.reprocess_batch(batch_id)

    def attach_notebooks_to_case(self, case_id, notebook_ids):
        current_notebooks = self.__api.case_list_attached_notebooks(case_id, results_only=True)
        current_notebooks = [notebook['id'] for notebook in current_notebooks]
        new_notebooks = helpers.format_notebook_ids(notebook_ids)
        updated_notebooks = current_notebooks + [x for x in new_notebooks if x not in current_notebooks]
        if current_notebooks == updated_notebooks:
            return {"result": "no changes made"}
        return self.__api.case_overwrite_attached_notebooks(case_id, updated_notebooks)

    def detach_notebooks_from_case(self, case_id, notebook_ids):
        current_notebooks = self.__api.case_list_attached_notebooks(case_id, results_only=True)
        if not current_notebooks:
            return {"result": "no changes made"}
        # Reformat to match
        current_notebooks = helpers.format_notebook_ids(current_notebooks)
        new_notebooks = helpers.format_notebook_ids(notebook_ids)
        updated_notebooks = [notebook for notebook in current_notebooks if notebook['id'] not in [x['id'] for x in new_notebooks]]
        if current_notebooks == updated_notebooks:
            return {"result": "no changes made"}
        return self.__api.case_overwrite_attached_notebooks(case_id, updated_notebooks)

    def remove_all_notebooks_from_case(self, case_id):
        current_notebooks = self.__api.case_list_attached_notebooks(case_id, results_only=True)
        if not current_notebooks:
            return {"result": "no changes made"}
        return self.__api.case_overwrite_attached_notebooks(case_id, [])

    # ToDo Add an action for basic search as well
    def search_alerts_advanced(self, query: str = None, limit: int = None, fetch_details=None, included_standard_fields=None, included_additional_fields=None):
        simple_format = False
        if fetch_details:
            if fetch_details == 'simple':
                simple_format = True
            elif fetch_details != 'standard':
                raise InputValidationError(input_var=fetch_details, action_description="fetch_details")
        query = query.strip() if query and query.strip() else ""
        result = self.__api.alerts_search_advanced(query=query, limit=limit)
        _ = self._result_dict_has_schema(result, "result", action_description="search alerts", raise_errors=True)
        output = result['result']
        if fetch_details:
            alerts = [x['id'] for x in output['data']]
            return [
                self.get_alert_by_id(alert, simple_format=simple_format, included_standard_fields=included_standard_fields, included_additional_fields=included_additional_fields)
                for alert in alerts
            ]
        return output

    # ToDo Expand on this. Offer a simple format, fetch details, etc.
    def search_cases_advanced(self, query: str = None, limit: int = None, **kwargs):
        # Validate the query first, just as the UI would do
        self.__api.cases_search_validate(query)

        result = self.__api.case_search_advanced(query=query, limit=limit, **kwargs)
        _ = self._result_dict_has_schema(result, "result", "data", action_description="search cases", raise_errors=True)
        return result["result"]["data"]

    def pause_stream(self, stream_id):
        result = self.__api.stream_pause(stream_id=stream_id)
        _ = self._result_dict_has_schema(result, "result", action_description="pause stream", raise_errors=True, accept_null=True)
        return result

    def resume_stream(self, stream_id):
        result = self.__api.stream_resume(stream_id=stream_id)
        _ = self._result_dict_has_schema(result, "result", action_description="resume stream", raise_errors=True, accept_null=True)
        return result

    def update_current_user_preferences(self, **kwargs):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise InputValidationError(
                input_var=kwargs, action_description="updating user preferences",
                message=f"Invalid input for updating user preferences. At least one ID with boolean value is required, but none were received."
            )
        for k, v in kwargs.items():
            if not isinstance(v, bool):
                raise InputValidationError(input_var=v, action_description="updating user preference", message=f"Invalid input for updating user preferences. Preference values must be boolean but received: {v}")
        preferences = self.__api.me()['result']['preferences']
        preference_ids = [p['id'] for p in preferences]
        for k in kwargs:
            if k not in preference_ids:
                raise InputValidationError(input_var=k, action_description="updating user preference", message=f"Invalid preference ID: {k}. Valid IDs are: {', '.join(preference_ids)}")
        for n in range(len(preferences)):
            if preferences[n]['id'] in kwargs:
                preferences[n]['value'] = kwargs[preferences[n]['id']]

        return self.__api.update_current_user_preferences(preferences)

    def create_user(self, username, email, authentication_type=None, group_names: list = None, group_ids: list = None):
        """
        Create a new LogicHub user

        :param username:
        :param email:
        :param authentication_type:
        :param group_names: ignored if group_ids is present
        :param group_ids:
        :return:
        """
        authentication_type = authentication_type or "password"
        if authentication_type != "password" and not isinstance(authentication_type, dict):
            raise InputValidationError(input_var=authentication_type, action_description="authentication type")
        if not group_ids:
            _groups = self.__api.formatted.user_groups_by_name
            if group_names:
                group_ids = []
                for g in group_names:
                    if g not in _groups:
                        raise UserGroupNotFound(input_var=g)
                    _group_id = helpers.format_user_group_id(_groups[g])
                    log.debug(f"Translated group {g} to ID {_group_id}")
                    group_ids.append(_group_id)
            else:
                group_ids = _groups["Everyone"]

        kwargs = {"username": username, "email": email, "authentication_type": authentication_type, "group_ids": group_ids}
        log.debug(f"Creating user: {json.dumps(kwargs)}")
        response = self.__api.create_user(**kwargs)
        return response['result']

    def delete_user_by_id(self, user_ids: Union[str, list]):
        if not isinstance(user_ids, list):
            user_ids = [user_ids]
        user_ids = [helpers.format_user_id(u) for u in user_ids]
        return self.__api.delete_user(user_ids=user_ids)

    def delete_user_by_name(self, usernames: Union[str, list]):
        _users = self.__api.formatted.users_by_name
        user_ids = []
        for u in (usernames if isinstance(usernames, list) else [usernames]):
            if u not in _users:
                log.error(f"User not found: {u}")
                raise UserNotFound(user=u)
            _user_id = helpers.format_user_id(_users[u])
            log.debug(f"Translated user {u} to ID {_user_id}")
            user_ids.append(_user_id)
        return self.delete_user_by_id(user_ids=user_ids)

    def get_password_settings(self) -> dict:
        result = self.__api.get_password_settings()
        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="get password settings")
        return result["result"]

    def update_password_settings(self, settings_dict):
        result = self.__api.update_password_settings(settings_dict=settings_dict)
        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="update password settings")
        return result["result"]


class LogicHub:
    verify_ssl = True
    http_timeout_default = LogicHubAPI.http_timeout_default

    def __init__(
            self, hostname, api_key=None, username=None, password=None,
            cache_seconds=None, verify_api_auth=True, default_timeout=None, logger: RootLogger = None, log_level=None, **kwargs):
        global log
        if logger:
            log = logger
        else:
            prep_generic_logger(level=log_level)
        self.hostname = hostname
        self.kwargs = kwargs
        self.api = LogicHubAPI(
            hostname=hostname, api_key=api_key, username=username, password=password,
            cache_seconds=cache_seconds, default_timeout=default_timeout or self.http_timeout_default,
            logger=logger, **kwargs)
        self.actions = Actions(self.api)
        if verify_api_auth:
            _ = self._verify_api_auth()
        log.debug(f"LogicHub (lhub) session successfully initialized (hostname={self.hostname})")

    def _verify_api_auth(self):
        _ = self.api.me()
        log.debug(f"Authentication successful (hostname={self.hostname})")
        return True



from setuptools import setup, find_packages

VERSION = '0.2.6.0'

setup(
    name="lhub",
    version=VERSION,
    author="Chad Roberts",
    author_email="chad@logichub.com",
    description="LogicHub API Wrapper",
    long_description="A Python package for interacting with LogicHub APIs",
    packages=find_packages(),
    install_requires=[
        "requests",
        "multipledispatch",
    ],
    keywords=["python", "lhub", "LogicHub", "api", "security"],

    url="https://github.com/deathbywedgie/lhub",
    project_urls={
        "Bug Tracker": "https://github.com/deathbywedgie/lhub/issues",
    },
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Natural Language :: English",
    ]
)



class LhBaseException(BaseException):
    """There was a generic exception that occurred while handling your
    request.

    Catching this exception will catch *all* custom exceptions from this package
    """
    message = "An exception occurred"

    def __init__(self, *args, **kwargs):
        self.input = kwargs.pop("input_var", None)
        if args:
            self.message = args[0]
        elif self.input:
            self.message += f": {self.input}"
        if not args:
            args = [self.message]

        self.url = kwargs.pop("url", None)
        self.last_response_status = kwargs.pop("last_response_status", None)
        self.last_response_text = kwargs.pop("last_response_status", None)
        super().__init__(*args, **kwargs)
